/*	libiac - Internal Application Communication Library
 *
 *	Authors: YE Nan <nan.ye@orange-ftgroup.com> 
 *	
 *	This software and associated documentation files (the "Software") 
 *	are copyright (C) 2005 LiPS Linux Phone Standards Forum [FranceTelecom] 
 *	All Rights Reserved. 
 *
 *	A copyright license is hereby granted for redistribution and use of 
 *	the Software in source and binary forms, with or without modification, 
 *	provided that the following conditions are met: 
 *	- Redistributions of source code must retain the above copyright notice, 
 *	this copyright license and the following disclaimer. 
 *  - Redistributions in binary form must reproduce the above copyright 
 * 	notice, this copyright license and the following disclaimer in the 
 *	documentation and/or other materials provided with the distribution. 
 *	- Neither the name of LiPS  nor the names of its Members may be used 
 *	to endorse or promote products derived from the Software without 
 *	specific prior written permission. 
 *
 *	A patent license for any Necessary Claims owned by Members of LiPS Forum 
 *	to make, have made, use, import, offer to sell, lease and sell or otherwise 
 *	distribute any implementation compliant with the any specification adopted 
 *	by the LiPS Forumcan be obtained from the respective Members on reasonable 
 *	and non-discriminatory terms and conditions and under reciprocity, as 
 *	regulated in more detail in the Internal Policy of the LiPS Forum. 
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDER, ITS MEMBERS AND CONTRIBUTORS 
 *	"AS IS", AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, 
 *	THE IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE 
 *	AND NON-INFRINGEMENT ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER, 
 *	ITS MEMBERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 *	SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 *	PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 *	OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 *	WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) 
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
 *	POSSIBILITY OF SUCH DAMAGE. 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <libintl.h>

#include <glib.h>

#include "iac.h"
#include "iaclock.h"
#include "iac-internal.h"
#include "iac-applauncher.h"

#include <dbus/dbus.h>

#define _(x) gettext(x)

#define	IAC_START_SERVICE_METHOD		"startservice_return"
#define	IAC_START_SERVICE_INTERFACE	"lips.iac.startservice"
#define	IAC_START_SERVICE_OBJ				"/lips/iac/startservice"

#define IAC_START_SERVICE_TIMEOUT		(10 * 1000)

static void
iac_dbus_message_print (DBusMessage * message)
{
	const gchar * service = NULL;
	const gchar * interface = NULL;
	const gchar * path = NULL;
	const gchar * method = NULL;

	service = dbus_message_get_destination(message);
	path = dbus_message_get_path(message);
	interface = dbus_message_get_interface(message);
	method = dbus_message_get_member(message);

	g_print("%s(): DBus-Message: 0x%08x\n", __FUNCTION__, (guint32)message);
	g_print("%s(): service     : %s\n", 
					__FUNCTION__,
					service ? service : "null");
	g_print("%s(): path        : %s\n",
					__FUNCTION__,
					path ? path : "null");
	g_print("%s(): interface   : %s\n",
					__FUNCTION__,
					interface ? interface : "null");
	g_print("%s(): method      : %s\n",
					__FUNCTION__,
					method ? method : "null");
	
	return;
}

static void
iac_context_printinfo (iac_context_t * context)
{
	g_print("%s(): context: 0x%0x8\n", __FUNCTION__, (guint32)context);
	if (context == NULL)
	{
		return;
	}
	
	g_print("%s(): |- service    : %s\n",
					 __FUNCTION__, 
					(context->service) ? (context->service) : "null");
	g_print("%s(): |- connection : 0x%08x\n",
					 __FUNCTION__,
					 (guint32)context->conn);
	g_print("%s(): `- listeners  : 0x%08x\n",
					__FUNCTION__,
					(guint32)context->listeners);
	
	return;
}

typedef struct
{
	iac_listener_func		 callback;
	gpointer						 data;
} iac_listener_t;

static void
message_attach_iac_val(DBusMessage			*message,
											 iac_value_t			*val)
{
		switch (val->vtype)
		{
			case IAC_TYPE_INT32:
//				g_print("type = int32   , value = %d\n", val->value.i32);
				dbus_message_append_args(message,
																 DBUS_TYPE_INT32,
																 &(val->value.i32),
																 DBUS_TYPE_INVALID);
				break;
					
			case IAC_TYPE_UINT32:
//				g_print("type = uint32  , value = %u\n", val->value.u32);
				dbus_message_append_args(message,
																 DBUS_TYPE_UINT32,
																 &(val->value.u32),
																 DBUS_TYPE_INVALID);
				break;
				
			case IAC_TYPE_BOOLEAN:
//				g_print("type = boolean , value = %s\n", val->value.b ? "true" : "false");
				dbus_message_append_args(message,
																 DBUS_TYPE_BOOLEAN,
																 &(val->value.b),
																 DBUS_TYPE_INVALID);
				break;
		  
				case IAC_TYPE_DOUBLE:
//				g_print("type = double  , value = %f\n", val->value.d);
				dbus_message_append_args(message,
																 DBUS_TYPE_DOUBLE,
																 &(val->value.d),
																 DBUS_TYPE_INVALID);
				break;
				
				case IAC_TYPE_STRING:
				if (val->value.s != NULL)
				{
//					g_print("type = string  , value = %s\n", val->value.s);
					dbus_message_append_args(message,
																	 DBUS_TYPE_STRING,
																	 &(val->value.s),
																	 DBUS_TYPE_INVALID);
				}
				break;
				
			default:
				break;	
		}
		
		return;
}

static void 
iac_val_parse_from_message (DBusMessageIter		*iter,
														iac_value_t				*val)
{
	int type = 0;
	
	type = dbus_message_iter_get_arg_type(iter);
	switch (type) {
		case DBUS_TYPE_INT32:
			val->vtype = IAC_TYPE_INT32;
			dbus_message_iter_get_basic(iter, &val->value.i32);
//			g_print("%s: val(int32)   : %d\n", __FUNCTION__, val->value.i32);
			break;
			
		case DBUS_TYPE_UINT32:
			val->vtype = IAC_TYPE_UINT32;
			dbus_message_iter_get_basic(iter, &val->value.u32);
//			g_print("%s: val(uint32)  : %u\n", __FUNCTION__, val->value.u32);
			break;
		
		case DBUS_TYPE_BOOLEAN:
			val->vtype = IAC_TYPE_BOOLEAN;
			dbus_message_iter_get_basic(iter, &val->value.b);
//			g_print("%s: val(boolean) : %s\n", __FUNCTION__, (val->value.b) ? "true" : "false");
			break;
  
		case DBUS_TYPE_DOUBLE:
			val->vtype = IAC_TYPE_DOUBLE;
			dbus_message_iter_get_basic(iter, &val->value.d);
//			g_print("%s: val(double)  : %f\n", __FUNCTION__, val->value.d);
			break;
		
		case DBUS_TYPE_STRING:
		{
			char * str = NULL;

			dbus_message_iter_get_basic(iter, &str);
			
			val->value.s = g_strdup (str);
			
			if(val->value.s == NULL) 
			{
				val->vtype = IAC_TYPE_INVALID;
			}
			else
			{
				val->vtype = IAC_TYPE_STRING;
			}
//			g_print("%s: val(string)  : %s\n", __FUNCTION__, val->value.s);
			break;
		}
		
		default:
//			g_print("%s: val(type%d)  : (unknown)\n", __FUNCTION__, val->vtype);
			val->vtype = IAC_TYPE_INVALID;
			val->value.i32 = 0;	    
			break;	
	}
	
	return;
}

static DBusHandlerResult
dbus_message_handler (DBusConnection	*connection,
											DBusMessage			*message,
											void						*user_data)
{
	iac_context_t * context = (iac_context_t *)user_data;
	iac_listener_t * listener = NULL;
	const gchar * interface = NULL;
	
	interface = dbus_message_get_interface(message);
	g_print("%s(): context = 0x%08x\n",
					__FUNCTION__,
					(guint32)context);
	g_print("%s(): interface = %s\n",
					__FUNCTION__,
					interface);
	
	listener = g_hash_table_lookup(context->listeners, interface);
	if (listener)
	{
		iac_arguments_t * args = NULL;
		DBusMessageIter iter;
		
		args = iac_arguments_new();
		dbus_message_iter_init(message, &iter);

    while (TRUE)
    {
			iac_value_t val;
				
			iac_val_parse_from_message(&iter, &val);
		
			g_array_append_val(args->values, val);
		
			if (dbus_message_iter_has_next(&iter))
			{
				dbus_message_iter_next(&iter);
			}
			else
			{
				break;
			}
		}

		if (TRUE)
		{
			iac_arguments_t * retval = NULL;
			iac_return_t ret = IAC_ERROR;
			
			retval = iac_arguments_new();
			if (retval == NULL)
			{
				return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
			}
			
//			g_print("%s(): calling listener callback\n", __FUNCTION__);
			ret = listener->callback(interface, 
															 dbus_message_get_member(message),
															 args,
															 retval,
															 listener->data);
															 
			g_print("%s(): ret = %s\n",
							__FUNCTION__,
							(ret == IAC_OK ? "IAC_OK" : "IAC_ERROR"));
			
			if(dbus_message_get_no_reply(message) == FALSE)			
			{
				DBusMessage * reply = NULL;
				
				if (ret == IAC_OK)
				{
					reply = dbus_message_new_method_return(message);
					if(reply != NULL) 
					{
						gint i;
//						iac_arguments_print(retval);
						for (i = 0; i < retval->values->len; i++)
						{
							iac_value_t * val;
							
							val = &g_array_index(retval->values, iac_value_t, i);
							message_attach_iac_val(reply, val);
						}
					}
					
					if (TRUE)
					{
						DBusMessageIter iter;
						
						dbus_message_iter_init(reply, &iter);
				
				    while (TRUE)
				    {
							if (dbus_message_iter_has_next(&iter))
							{
								dbus_message_iter_next(&iter);
							}
							else
							{
								break;
							}
						}
					}
				}
				else
				{
					gchar * error_name = NULL;
					
					error_name = g_strdup_printf("%s.%s.%s.error",
																				dbus_message_get_destination(message),
																				dbus_message_get_interface(message),
																				dbus_message_get_member(message));
					g_print("%s(): listener return error: %s\n",
									__FUNCTION__,
									error_name);
					reply = dbus_message_new_error(message, error_name, "iac_error");
					g_free(error_name);
				}
				
				if (reply != NULL)
				{
//					g_print("%s(): sending reply\n", __FUNCTION__);
					dbus_connection_send(context->conn,
															 reply,
															 NULL);
															 
					dbus_connection_flush(context->conn);
					dbus_message_unref(reply);
				}
			}
			
			iac_arguments_free(retval);
		}
		
		iac_arguments_free(args);
 											 
		return DBUS_HANDLER_RESULT_HANDLED;
	}
	
  return DBUS_HANDLER_RESULT_NOT_YET_HANDLED;
}

iac_return_t
iac_register_listener (iac_context_t				*context,
											 const gchar					*object,
											 const gchar					*interface,
											 iac_listener_func		 callback,
											 gpointer							 user_data)
{
	iac_listener_t * listener = NULL;
	struct DBusObjectPathVTable vt;
	gchar * key = NULL;

	vt.message_function			= dbus_message_handler;
	vt.unregister_function	= NULL;
	g_print("%s(): register listener:\n", __FUNCTION__);
	g_print("%s(): path      : %s\n", __FUNCTION__, object);
	g_print("%s(): interface : %s\n", __FUNCTION__, interface);
	g_print("%s(): callback  : 0x%08x\n", __FUNCTION__, (guint32)callback);
	g_print("%s(): callback  : 0x%08x\n", __FUNCTION__, (guint32)user_data);
	if (dbus_connection_register_object_path(context->conn,
																					 object,
																					 &vt,
																					 (gpointer)context) == FALSE)
	{
/*
		g_print("%s(): failed to register vtable for %s\n",
						__FUNCTION__,
						path);
						
		return FALSE;
*/
	}																								 
	
	listener = g_new0(iac_listener_t, 1);
	if (listener == NULL)
	{
		return IAC_ERROR;
	}
	listener->callback	= callback;
	listener->data			= user_data;
	
	key = g_strdup(interface);

	if (context->listeners == NULL)
	{
		context->listeners	= g_hash_table_new_full(g_str_hash,
	  																						g_str_equal,
	  																						g_free,
	  																						g_free);
	}  

	g_hash_table_replace(context->listeners,
											 key,
											 (gpointer)listener);										
	
	return IAC_OK;
}

iac_return_t
iac_unregister_listener (iac_context_t				*context,
												 const gchar					*object,
												 const gchar					*interface)
{
	g_hash_table_remove(context->listeners,
											interface);

/*	
	dbus_connection_unregister_object_path(context->conn,
																				 path);
*/
	return IAC_OK;
}

typedef struct {
	GMainLoop			*loop;
	gint					 ss_ret;
} iac_ss_t;

static iac_ss_t *
iac_ss_new (void)
{
	iac_ss_t * ss = NULL;
	
	ss = g_new0(iac_ss_t, 1);
	if (ss == NULL)
	{
		return NULL;
	}
	
	ss->loop = g_main_loop_new(NULL, FALSE);
	if (ss->loop == NULL)
	{
		g_free(ss);
		return NULL;
	}
	
	ss->ss_ret	= IAC_ERROR;
	
	return ss;
}

static void
iac_ss_free (iac_ss_t * ss)
{
	if (ss == NULL)
	{
		return;
	}
	
	if (ss->loop)
	{
		g_main_loop_unref(ss->loop);
	}
	
	g_free(ss);
		
	return;
}

static iac_return_t	
iac_start_service_handler (const gchar				*interface,
													 const gchar				*method,
													 iac_arguments_t		*arguments,
													 iac_arguments_t		*retval,
													 gpointer						 user_data)
{
	iac_ss_t * ss = (iac_ss_t *)user_data;
	
	g_print("%s(): interface: %s\n", __FUNCTION__, interface);
	g_print("%s(): method   : %s\n", __FUNCTION__, method);
	g_print("%s(): arguments: 0x%08x\n", __FUNCTION__, (guint32)arguments);
	
	if (g_strcasecmp(interface, IAC_START_SERVICE_INTERFACE) == 0 &&
			g_strcasecmp(method, IAC_START_SERVICE_METHOD) == 0)
	{
		iac_arguments_get_args(arguments,
													 IAC_TYPE_INT32,
													 &(ss->ss_ret),
													 IAC_TYPE_INVALID);
		g_print("%s(): startservice return = %d\n",
						__FUNCTION__,
						ss->ss_ret);

		g_print("%s(): quit looping (0x%08x)\n",
						__FUNCTION__,
						(guint32)ss->loop);
		g_main_loop_quit(ss->loop);
		
		return IAC_OK;
	}

	return IAC_ERROR;
}

static gboolean
iac_start_service_timeout_cb (gpointer data)
{
	iac_ss_t * ss = (iac_ss_t *)data;
	
	g_main_loop_quit(ss->loop);
	
	return TRUE;
}

iac_return_t
iac_start_service (iac_context_t		*context,
									 const gchar			*service,
									 const gchar			*method)
{
	DBusMessage * message = NULL;
	DBusMessage * reply;
	DBusError error;
	const gchar * unique_name = NULL;
	gint ss_ret = IAC_ERROR;
	
	if (iac_lock() == FALSE)
	{
		g_print("%s(): is locked\n", __FUNCTION__);
		return IAC_ERROR;
	}
	
	unique_name = dbus_bus_get_unique_name(context->conn);
	g_print("%s(): unique name = %s\n",
					__FUNCTION__,
					unique_name ? unique_name : "null");
	
	message = dbus_message_new_method_call(IAC_APP_LAUNCHER_SERVICE,
																				 IAC_APP_LAUNCHER_OBJECT,
																				 IAC_APP_LAUNCHER_INTERFACE,
																				 IAC_APP_LAUNCHER_METHOD);

	dbus_message_append_args(message,
													 DBUS_TYPE_STRING,
													 &service,
													 DBUS_TYPE_STRING,
													 &method,
													 DBUS_TYPE_STRING,
													 &unique_name,
													 DBUS_TYPE_INVALID);
	
//	iac_dbus_message_print(message);													 

	dbus_error_init(&error);
		
	reply = dbus_connection_send_with_reply_and_block(context->conn,
																										message, 
																										context->timeout,
																										&error);
	if (reply == NULL)
	{
		g_print("%s(): reply error: %s\n", 
						__FUNCTION__,
						error.message);

		dbus_message_unref(message);
    dbus_error_free(&error);

		iac_unlock();
		
		return IAC_ERROR;
	}
	else
	{
		int ret = IAC_APP_LAUNCHE_FAILURE;
		DBusError error;
		
		dbus_error_init(&error);
		
		if (dbus_message_get_args(reply,
															&error,
															DBUS_TYPE_INT32,
															&ret,
															DBUS_TYPE_INVALID) == FALSE)
		{
			g_print("%s(): reply error: %s\n", 
							__FUNCTION__,
							error.message);
	
			dbus_message_unref(message);
			dbus_message_unref(reply);
	    dbus_error_free(&error);
		    
			iac_unlock();

			return IAC_ERROR;
		}
													
		if (ret != IAC_APP_LAUNCHE_SUCCESS)
		{
			iac_unlock();

			return IAC_ERROR;
		}															
	}
															 
	dbus_message_unref(message);	
	dbus_message_unref(reply);
	
	if (TRUE)
	{
		iac_ss_t * ss = NULL;
		gint timer;
		
		ss = iac_ss_new();
		if (ss == NULL)
		{
			iac_ss_free(ss);
			iac_unlock();

			return IAC_ERROR;
		}
		
		timer = g_timeout_add(IAC_START_SERVICE_TIMEOUT,
													iac_start_service_timeout_cb,
													ss);
		g_print("%s(): add timer(%d) for %d\n",
						__FUNCTION__,
						timer,
						IAC_START_SERVICE_TIMEOUT);
		
		iac_register_listener(context,
													IAC_START_SERVICE_OBJ,
													IAC_START_SERVICE_INTERFACE,
													iac_start_service_handler,
													ss);
		
		g_print("%s(): entering loop (0x%08x)\n",
						__FUNCTION__,
						(guint32)ss->loop);

		g_main_loop_run(ss->loop);
		
		g_print("%s(): leaved loop (0x%08x)\n",
						__FUNCTION__,
						(guint32)ss->loop);
		
		g_source_remove(timer);
		
		ss_ret = ss->ss_ret;
	
		iac_ss_free(ss);
	}
	
	iac_unlock();
	
	return ss_ret;
}

iac_return_t iac_execute_service (iac_context_t				*context,
																	const gchar					*service,
																	const gchar					*object,
																	const gchar					*interface,
																	const gchar					*method,
																	iac_arguments_t			*args,
																	iac_arguments_t			*retval)
{
	DBusMessage * message = NULL;
	
	message = dbus_message_new_method_call(service,
																				 object,
																				 interface,
																				 method);
	
	if (args)
	{
		gint i;
		for (i = 0; i < args->values->len; i++)
		{
			iac_value_t * val;
			
			val = &g_array_index(args->values, iac_value_t, i);
			message_attach_iac_val(message, val);
		}
	}
	
	if (retval == NULL)
	{
		iac_dbus_message_print(message);
		
		dbus_message_set_no_reply(message, TRUE);
		if (dbus_connection_send(context->conn,
														 message,
														 NULL) == FALSE)
		{
			dbus_message_unref(message);
			return IAC_ERROR;
		}
		
		dbus_connection_flush(context->conn);
//		g_print("%s(): iac send with no-reply\n", __FUNCTION__);
	}
	else
	{
		DBusMessage * reply;
		DBusError error;

		dbus_error_init(&error);
		
//		g_print("%s(): iac send and block for reply\n", __FUNCTION__);
		reply = dbus_connection_send_with_reply_and_block(context->conn,
																											message, 
																											context->timeout,
																											&error);

		if (reply == NULL)
		{
			g_print("%s(): replay error: %s\n", 
							__FUNCTION__,
							error.message);

			dbus_message_unref(message);
	    dbus_error_free(&error);
	    
			return IAC_ERROR;
		}
		else
		{
			switch(dbus_message_get_type(reply)) 
			{
	      case DBUS_MESSAGE_TYPE_ERROR:
	      {
	      	DBusError error;
	      	iac_value_t val;
	      	
					dbus_set_error_from_message(&error, reply);
					val.vtype = IAC_TYPE_STRING;
					val.value.s = g_strdup(error.message);
					g_array_append_val(retval->values, val);
					dbus_error_free(&error);

					dbus_message_unref(reply);
					dbus_message_unref(message);

					return IAC_ERROR;
	      }
	      
	      case DBUS_MESSAGE_TYPE_METHOD_RETURN:
	      {
					DBusMessageIter iter;
					
					dbus_message_iter_init(reply, &iter);
			
			    while (TRUE)
			    {
						iac_value_t val;
							
						iac_val_parse_from_message(&iter, &val);
					
						g_array_append_val(retval->values, val);
					
						if (dbus_message_iter_has_next(&iter))
						{
							dbus_message_iter_next(&iter);
						}
						else
						{
							break;
						}
					}
					break;
	      }
	      
	      default:
	      {
	      	iac_value_t val;
	      	
	      	val.vtype = IAC_TYPE_STRING;
	      	val.value.s = g_strdup("Unkonwn return");
					g_array_append_val(retval->values, val);
				
					dbus_message_unref(reply);
					dbus_message_unref(message);
					
					return IAC_ERROR;
	      }
			}
		}
	}
															 
	dbus_message_unref(message);
	
	return IAC_OK;
}

typedef struct {
	gchar								*interface;
	gchar								*method;
	iac_async_callback	 func;
	void								*user_data;
} iac_async_cb_t;

static void
iac_async_cb_free (iac_async_cb_t * async)
{
	if (async == NULL)
	{
		return;
	}
	
	if (async->interface)
	{
		g_free(async->interface);
	}

	if (async->method)
	{
		g_free(async->method);
	}
	
	g_free(async);
	
	return;
}

static void
dbus_async_pending_handler (DBusPendingCall		*pending,
														void							*user_data)
{
	iac_async_cb_t * async = (iac_async_cb_t *)user_data;
	DBusMessage * message = NULL;
	gint type = DBUS_MESSAGE_TYPE_ERROR;
	
	message = dbus_pending_call_steal_reply(pending);
	if (message == NULL)
	{
		return;
	}
	
	type = dbus_message_get_type(message);
	switch (type)
	{
		case DBUS_MESSAGE_TYPE_METHOD_RETURN:
		{
			iac_arguments_t * retval = NULL;
			DBusMessageIter iter;
			
			iac_dbus_message_print(message);
			
			dbus_message_iter_init(message, &iter);
			retval = iac_arguments_new();
			if (retval == NULL)
			{
				dbus_message_unref(message);
			}
			
			while (TRUE)
			{
				iac_value_t val;
							
				iac_val_parse_from_message(&iter, &val);
					
				g_array_append_val(retval->values, val);
					
				if (dbus_message_iter_has_next(&iter))
				{
					dbus_message_iter_next(&iter);
				}
				else
				{
					break;
				}
			}
			
			async->func(async->interface,
									async->method,
									retval,
									async->user_data);
									
			iac_arguments_free(retval);
			
			break;
		}
		
		case DBUS_MESSAGE_TYPE_ERROR:
			g_print("%s(): return error\n", __FUNCTION__);
			break;
		
		default:
			g_print("%s(): return unkown\n", __FUNCTION__);
			break;
	}		
	
	return;
}


iac_return_t iac_execute_service_async (iac_context_t				*context,
//																				const gchar					*app_service,
																				const gchar					*service,
																				const gchar					*object,
																				const gchar					*interface,
																				const gchar					*method,
																				iac_arguments_t			*args,
																				iac_async_callback	 async_cb,
																				void								*user_data)
{
	DBusMessage * message = NULL;
	
	message = dbus_message_new_method_call(service,
																				 object,
																				 interface,
																				 method);
	
	if (args)
	{
		gint i;
		for (i = 0; i < args->values->len; i++)
		{
			iac_value_t * val;
			
			val = &g_array_index(args->values, iac_value_t, i);
			message_attach_iac_val(message, val);
		}
	}
	
	if (async_cb == NULL)
	{
		g_print("%s(): async_cb is null", __FUNCTION__);

		dbus_message_set_no_reply(message, TRUE);
		if (dbus_connection_send(context->conn,
														 message,
														 NULL) == FALSE)
		{
			dbus_message_unref(message);
			return IAC_ERROR;
		}
		
		dbus_connection_flush(context->conn);
//		g_print("%s(): iac send with no-reply\n", __FUNCTION__);
	}
	else
	{
    DBusPendingCall * pending = NULL;

		if (dbus_connection_send_with_reply(context->conn,
																				message,
																				&pending,
																				context->timeout) != TRUE)
		{
			g_print("%s(): dbus_connection_send_with_reply error\n", 
							__FUNCTION__);

			dbus_message_unref(message);
	    
			return IAC_ERROR;
		}
		else
		{
			iac_async_cb_t * asyn = NULL;
			
			asyn = g_new(iac_async_cb_t, 1);
			if (asyn == NULL)
			{
				dbus_message_unref(message);
				return IAC_ERROR;
			}
			g_print("%s(): setting pending call\n", __FUNCTION__);
			asyn->func			= async_cb;
			asyn->interface = g_strdup(interface);
			asyn->method		= g_strdup(method);
			asyn->user_data	= user_data;
			
			dbus_pending_call_set_notify(pending,
																	 dbus_async_pending_handler,
																	 asyn,
																	 (DBusFreeFunction)iac_async_cb_free);
		}
		
		dbus_connection_flush(context->conn);
	}
															 
	dbus_message_unref(message);
	
	return IAC_OK;
}

iac_return_t iac_emit_signal (iac_context_t				*context,
															const gchar					*object,
															const gchar					*interface,
															const gchar					*method,
															iac_arguments_t			*args)
{
	DBusMessage * message = NULL;
	
	message = dbus_message_new_signal(object,
																		interface,
																		method);
	
	if (args)
	{
		gint i;
		for (i = 0; i < args->values->len; i++)
		{
			iac_value_t * val;
			
			val = &g_array_index(args->values, iac_value_t, i);
			message_attach_iac_val(message, val);
		}
	}
	
	iac_dbus_message_print(message);
		
	dbus_message_set_no_reply(message, TRUE);
	
	if (dbus_connection_send(context->conn,
													 message,
													 NULL) == FALSE)
	{
		dbus_message_unref(message);
		return IAC_ERROR;
	}
		
	dbus_connection_flush(context->conn);
															 
	dbus_message_unref(message);
	
	return IAC_OK;
}
																										 
iac_context_t *
iac_initialize (void)
{
	iac_context_t * context = NULL;
	DBusConnection * conn = NULL;
  DBusError error;

  dbus_error_init (&error);

  conn = dbus_bus_get(DBUS_BUS_SESSION, &error);
	if (!conn) 
	{
		g_print("%s(): failed to connect to the D-BUS daemon, %s\n",
						__FUNCTION__,
						error.message);
		dbus_error_free (&error);
		
		return NULL;
  }
  
  context = g_new0(iac_context_t, 1);
  if (context == NULL)
  {
  	return NULL;
  }

	dbus_connection_setup_with_g_main(conn, NULL);

  context->conn				= conn;
  context->service		= NULL;
  context->listeners	= NULL;
  context->timeout		= -1;

  return context;
}

iac_return_t
iac_register_service (iac_context_t		*context,
											const gchar			*service)
{
  DBusError error;

  dbus_error_init (&error);
  
  dbus_bus_request_name(context->conn, service, 0, &error);
  if (dbus_error_is_set(&error)) 
  {
		g_print("%s(): failed to acquire service %s, %s\n",
						__FUNCTION__,
						service,
						error.message);
    dbus_error_free (&error);
    
    return IAC_ERROR;
  }

	context->service		= g_strdup(service);
	
  return IAC_OK;
}

void
iac_deinitialize (iac_context_t * context)
{
	if (context == NULL)
	{
		return;
	}
	
	if (context->service)
	{
		g_free(context->service);
	}
	
	if (context->listeners)
	{
		g_hash_table_destroy(context->listeners);
	}

	g_free(context);
	
	return;
}

/*
vi:ts=2:nowrap:ai:expandtab 
*/
